{*************************************************************************}
{*   Configs                                                             *}
{*   Copyright (C) 2004 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\Configs.pas                                            *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 1st March 2004  (01.03.2004)                               }
{*************************************************************************}

{This unit contains classes for storage and management of configuration data.}

unit Configs;

{ TODOS:
  - as in CS's cfg ...  System.VFS = vfs.cfg
  - comments stay in saved configs
  - implement XML config driver
  - it is persistent so add streaming
}

{$I Compiler.inc}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Files, Lists, DOM;

const
  {Char used as delimeter of keys in config path. Ex: Video.Resolution.Width
   (where '.' is sConfigPathDelim).}
  sConfigPathDelim = '.';

  {Default extension of text config files.}
  sTextConfigExt = 'cfg';
  {Char used in text config files to separate keys from values. Ex:
   Video.Resolution.Width=800 (where '=' is delim).}
  sTextConfigValueDelim = '=';
  {Chars used to mark line in text config file as comment.}
  TextConfigCommnets = ['/', ';', '{'];

type

  {Defines format in which TNConfigFile will be saved:
    csText - config will be saved as text file.
    csXML  - config will be saved as XML document.}
  TConfigStorage = (csText, csXML);

  {Mode in which TNConfigDriver works.}
  TDriverMode = (dmRead, dmWrite);

  {Abstract class used by TNConfigFile to do real saving/loading
   of config data.}
  TNAbstractConfigDriver = class(TNSingleton)
  protected
    FStream: TNStream;
    FMode: TDriverMode;
  public
    {Creates new config driver using AStream in AMode mode.}
    constructor Create(AStream: TNStream; AMode: TDriverMode); reintroduce;
      overload;
    destructor Destroy; override;
    {Starts new section called NodeName.}
    procedure BeginNode(const NodeName: string); virtual; abstract;
    {Ends current section.}
    procedure EndNode; virtual; abstract;
    {Reads next key and value from stream and moves position in
     source data to next key+value.}
    procedure ReadValue(var Key, Value: string); virtual; abstract;
    {Writes Key and Value to actual position in stream.}
    procedure WriteValue(const Key, Value: string); virtual; abstract;
    {Returns True if there is no further config data in stream.}
    function IsEnd: Boolean; virtual; abstract;
    {In dmRead mode this is called in constructor, in dmWrite mode
     it is called in destructor. It can be used to preload/writeall
     data from/to stream.}
    procedure Flush; virtual; abstract;
  end;

  {Class refernce for TNAbstractConfigDriver.}
  TNConfigDriverClass = class of TNAbstractConfigDriver;

  {Config driver operating on text files. Files must be in
   'Key1.Key2.Key3.Key4=Value' format.
   Comments are allowed: they must start with one of TextConfigCommnets
   chars or simply do not containt '=' character.
   Line is broken to Key and Value at position of first '=' char
   so Value can contain additional '=' chars.
   All leading and trailing spaces and control characters are trimmed from
   string (by SysUtils.Trim()).
   Warning: input stream is read from actual pos to end in dmRead mode
   so this driver is not "stream safe".
   Example:
     Video.Resolution.Width=1024
     // this is config comment
     Games.RGP.Best = Daggerfall
     Character.Name=Tiber Septim}
  TNTextConfigDriver = class(TNAbstractConfigDriver)
  protected
    FParent: string;
    FBuffer: string;
  public
    procedure BeginNode(const NodeName: string); override;
    procedure EndNode; override;
    procedure ReadValue(var Key, Value: string); override;
    procedure WriteValue(const Key, Value: string); override;
    function IsEnd: Boolean; override;
    procedure Flush; override;
  end;

  {Config driver operating on XML documents.}
  TNXMLConfigDriver = class(TNAbstractConfigDriver)
  protected
  public
  end;

  {Class for storage configuration data. Each TNConfigFile can have
   other TNConfigFiles as children so large hierarchies can be
   created. All values are stored as strings here.
   If you want to access some value in config file you must use
   string called Key. Each Key can contain more subkeys separated
   by sConfigPathDelim (usualy '.'). Each subkey refers to
   TNConfigFile deeper in hierarchy. Last subkey in Key is
   name of value owned by TNConfigFile refered by previous subkey.
   So if you have Key 'Raiders.Germans.Vizigoths' then
   'Raiders' is child (node) of current config. 'Raiders' has
   child called 'Germans' and 'Germans' has value called 'Vizigoths'.}
  TNConfigFile = class(TNNamedPersistent)
  protected
    FNodes: TNSortedStringObjects;
    FValues: TNSortedDictionary;
    FStorage: TConfigStorage;
    function GetNode(Index: Long): TNConfigFile;
    function GetNodeCount: Long;
    function GetValueCount: Long;
    function GetValue(Index: Long): string;
    procedure SetValue(Index: Long; const Value: string);
    function GetValueKey(Index: Long): string;
    {Adds new node called NodeName.}
    procedure AddNode(const NodeName: string);
    {Saves this node using given config driver.}
    procedure SaveAsNode(Driver: TNAbstractConfigDriver);
  public
    constructor Create; overload; override;
    {Calls Create() constructor and then calls LoadFromFile(AFileName).
     AStorage identifies in which format is AFileName config.}
    constructor Create(const AFileName: string;
      AStorage: TConfigStorage = csText); reintroduce; overload;
    destructor Destroy; override;

    {Reads string value and if successful returns it. Otherwise returns Default.}
    function ReadString(const Key: string; const Default: string = ''): string;
    {Reads boolean value and if successful returns it. Value can be stored
     in config either as number (0,1) or string (True, False). On failure returns
     Default.}
    function ReadBool(const Key: string; Default: Boolean = False): Boolean;
    {Reads integer value and if successful returns it. Otherwise returns Default.}
    function ReadInt(const Key: string; const Default: Int64 = 0): Int64;
    {Reads float value and if successful returns it. Otherwise returns Default.}
    function ReadFloat(const Key: string; const Default: Float = 0.0): Float;

    {Writes string value to config.}
    procedure WriteString(const Key, Value: string);
    {Writes boolean value to config.}
    procedure WriteBool(const Key: string; Value: Boolean);
    {Writes integer value to config.}
    procedure WriteInt(const Key: string; const Value: Int64);
    {Writes float value to config.}
    procedure WriteFloat(const Key: string; const Value: Float);

    {Adds given Config as node of current TNConfigFile. Config's
     name (path used in Key) is replaced by NewName.
     This file adds reference to Config to nodes it does not create
     copy of Config (so do not Destroy Config after Add).
     If you want to copy contents of Config use AddConfig(Config.Clone)
     instead.}
    procedure AddConfig(const NewName: string; Config: TNConfigFile);
    {Copies all values and nodes from Config (all nodes are cloned).}
    procedure CopyFromConfig(Config: TNConfigFile);
    {Creates new instance of TNConfigFile and copies all values and nodes
     to it.}
    function Clone: TNConfigFile;
    {Deletes all configuration data of this node.}
    procedure Clear;
    {Returns subnode in node hierarchy defined by NodeName
     (it can contain subkeys). If node is not found on given path nil
     is returned.}
    function FindSubNode(const NodeName: string): TNConfigFile;

    {Loads new configuration data from file. Current data are deleted.}
    procedure LoadFromFile(const FileName: string);
    {Loads new configuration data from stream. Current data are deleted.}
    procedure LoadFromStream(Stream: TNStream);
    {Saves actual configuration data to file.}
    procedure SaveToFile(const FileName: string);
    {Saves actual configuration data to stream.}
    procedure SaveToStream(Stream: TNStream);

    {Returns number of this config's nodes.}
    property NodeCount: Long read GetNodeCount;
    {Returns node at given index.}
    property Nodes[Index: Long]: TNConfigFile read GetNode;
    {Returns number of valus of this config.}
    property ValueCount: Long read GetValueCount;
    {Returns Key of value at given index.}
    property ValueKeys[Index: Long]: string read GetValueKey;
    {Returns value at given index as string (all values as internally
     stored as strings).}
    property Values[Index: Long]: string read GetValue write SetValue;
    {Defines how this config saves/loads its data. Note that
     if you load config saved with different Storage it will fail or crash.}
    property Storage: TConfigStorage read FStorage write FStorage;
  end;

{Returns class reference of config driver defined by given Storage.}
function GetDriverClass(Storage: TConfigStorage): TNConfigDriverClass;


{*************************************************************************}
{                             IMPLEMEMTATION                              }
{*************************************************************************}


implementation


function GetDriverClass(Storage: TConfigStorage): TNConfigDriverClass;
begin
  Result := nil;
  case Storage of
    csText: Result := TNTextConfigDriver;
    csXML: Result := TNXMLConfigDriver;
  end;
end;


{*************************************************************************}
{                          TNConfigDriver methods                         }
{*************************************************************************}


constructor TNAbstractConfigDriver.Create(AStream: TNStream; AMode:
  TDriverMode);
begin
  Create;
  FMode := AMode;
  FStream := AStream;
  if FMode = dmRead then
    Flush;
end;

destructor TNAbstractConfigDriver.Destroy;
begin
  if FMode = dmWrite then
    Flush;
  inherited Destroy;
end;


{*************************************************************************}
{                          TNTextConfigDriver methods                     }
{*************************************************************************}


procedure TNTextConfigDriver.BeginNode(const NodeName: string);
begin
  if (NodeName <> '') and (FParent <> '') then
    FParent := FParent + sConfigPathDelim + NodeName
  else
    FParent := NodeName;
end;

procedure TNTextConfigDriver.EndNode;
begin
  StrTokenEnd(FParent, sConfigPathDelim);
end;

procedure TNTextConfigDriver.ReadValue(var Key, Value: string);
var
  Start, P: PChar;
  I: Long;
  S: string;
begin
  Key := '';
  Value := '';
  P := Pointer(FBuffer);
  if P <> nil then
  begin
    Start := P;
    // inc P to get to end of line
    while not (P^ in [#0, #10, #13]) do
      Inc(P);
    SetString(S, Start, P - Start);
    // inc P to get to start of new line
    while (P^ in [#10, #13]) do
      Inc(P);
    Delete(FBuffer, 1, P - Start);
    // check if S is valid line of type Key = Value
    if S <> '' then
    begin
      I := PosNoCase(sTextConfigValueDelim, S);
      if not ((S[1] in TextConfigCommnets) or (I = 0)) then
      begin
        Key := Trim(Copy(S, 1, I - 1));
        Value := Trim(Copy(S, I + 1, MaxLong));
      end;
    end;
  end;
end;

procedure TNTextConfigDriver.WriteValue(const Key, Value: string);
var
  S: string;
begin
  if FParent <> '' then
    S := FParent + sConfigPathDelim + Key
  else
    S := Key;
  FBuffer := FBuffer + S + sTextConfigValueDelim + Value + sLineBreak;
end;

function TNTextConfigDriver.IsEnd: Boolean;
begin
  Result := FBuffer = '';
end;

procedure TNTextConfigDriver.Flush;
var
  Size: Long;
begin
  case FMode of
    dmWrite:
      begin
        FStream.Write(FBuffer[1], Length(FBuffer));
        FBuffer := '';
      end;
    dmRead:
      begin
        Size := FStream.Size - FStream.Position;
        SetString(FBuffer, nil, Size);
        FStream.Read(FBuffer[1], Size);
      end;
  end;
end;


{*************************************************************************}
{                           TNConfigFile methods                          }
{*************************************************************************}


constructor TNConfigFile.Create;
begin
  inherited Create;
  FNodes := TNSortedStringObjects.Create;
  FValues := TNSortedDictionary.Create;
  FName := '';
end;

constructor TNConfigFile.Create(const AFileName: string;
  AStorage: TConfigStorage);
begin
  Create;
  FStorage := AStorage;
  LoadFromFile(AFileName);
end;

destructor TNConfigFile.Destroy;
begin
  Clear;
  FNodes.Free;
  FValues.Free;
  inherited Destroy;
end;

function TNConfigFile.GetNodeCount: Long;
begin
  Result := FNodes.Count;
end;

function TNConfigFile.GetNode(Index: Long): TNConfigFile;
begin
  Result := TNConfigFile(FNodes.Objects[Index]);
end;

function TNConfigFile.GetValueCount: Long;
begin
  Result := FValues.Count;
end;

function TNConfigFile.GetValue(Index: Long): string;
begin
  Result := FValues.Words[Index];
end;

procedure TNConfigFile.SetValue(Index: Long; const Value: string);
begin
  FValues.Words[Index] := Value;
end;

function TNConfigFile.GetValueKey(Index: Long): string;
begin
  Result := FValues.Keys[Index];
end;

procedure TNConfigFile.AddNode(const NodeName: string);
var
  Node: TNConfigFile;
begin
  Node := TNConfigFile.Create;
  Node.Name := NodeName;
  FNodes.Add(NodeName, Node);
end;

procedure TNConfigFile.SaveAsNode(Driver: TNAbstractConfigDriver);
var
  I: Long;
begin
  Driver.BeginNode(Name);
  for I := 0 to FValues.Count - 1 do
    Driver.WriteValue(FValues.Keys[I], FValues.Words[I]);
  for I := 0 to FNodes.Count - 1 do
    TNConfigFile(FNodes.Objects[I]).SaveAsNode(Driver);
  Driver.EndNode;
end;

function TNConfigFile.ReadString(const Key, Default: string): string;
var
  I: Long;
  SubKey, NodeName: string;
begin
  Result := Default;
  I := PosNoCase(sConfigPathDelim, Key);
  if I > 0 then
  begin
    // search for value in subnodes
    NodeName := Copy(Key, 1, I - 1);
    SubKey := Copy(Key, I + 1, MaxLong);
    if FNodes.Find(NodeName, I) then
      Result := TNConfigFile(FNodes.Objects[I]).ReadString(SubKey, Default);
  end
  else
    // value should be in this TNConfigFile instance
    if FValues.Find(Key, I) then
      Result := FValues.Words[I];
end;

function TNConfigFile.ReadBool(const Key: string; Default: Boolean): Boolean;
var
  S: string;
  I: Long;
begin
  S := ReadString(Key);
  Result := TryStrToInt(S, I);
  if Result then
    Result := I <> 0
  else
  begin
    Result := SameStr(S, sTrue);
    if not Result then
    begin
      Result := not SameStr(S, sFalse);
      if Result then
        Result := Default;
    end;
  end;
end;

function TNConfigFile.ReadInt(const Key: string; const Default: Int64): Int64;
var
  S: string;
begin
  S := ReadString(Key);
  if (Length(S) > 2) and (S[1] = '0') and ((S[2] = 'X') or (S[2] = 'x')) then
    S := '$' + Copy(S, 3, MaxLong);
  Result := StrToInt64Def(S, Default);
end;

function TNConfigFile.ReadFloat(const Key: string; const Default: Float): Float;
begin
  Result := StrToFloatDef(ReadString(Key), Default);
end;

procedure TNConfigFile.WriteString(const Key, Value: string);
var
  I: Long;
  SubKey, NodeName: string;
begin
  if Key <> '' then
  begin
    I := PosNoCase(sConfigPathDelim, Key);
    if I > 0 then
    begin
      // Value belongs to some subnode
      NodeName := Copy(Key, 1, I - 1);
      SubKey := Copy(Key, I + 1, MaxLong);
      if FNodes.Find(NodeName, I) then
      begin
        // write string to existing subnode
        TNConfigFile(FNodes.Objects[I]).WriteString(SubKey, Value);
      end
      else
      begin
        // node in key path does not exist so create one and write
        AddNode(NodeName);
        if FNodes.Find(NodeName, I) then
          TNConfigFile(FNodes.Objects[I]).WriteString(SubKey, Value);
      end;
    end
    else
    begin
      // Value belongs to this TNConfigFile instance
      if FValues.Find(Key, I) then
        // value already exists so overwrite it
        FValues.Words[I] := Value
      else
        // value does not exist so create it
        FValues.Add(Key, Value);
    end;
  end;
end;

procedure TNConfigFile.WriteBool(const Key: string; Value: Boolean);
begin
  WriteString(Key, Iff(Value, sTrue, sFalse));
end;

procedure TNConfigFile.WriteInt(const Key: string; const Value: Int64);
begin
  WriteString(Key, IntToStr(Value));
end;

procedure TNConfigFile.WriteFloat(const Key: string; const Value: Float);
begin
{$IFDEF FLOATSINGLE}
  WriteString(Key, Format('%.4f', [Value]));
{$ELSE}
  WriteString(Key, Format('%.10f', [Value]));
{$ENDIF}
end;

procedure TNConfigFile.AddConfig(const NewName: string; Config: TNConfigFile);
begin
  if Config <> nil then
  begin
    Config.Name := NewName;
    if Config.Name = '' then
      Config.Name := IntToStr(Long(Config));
    // adds Config to nodes so do not Free it now!
    FNodes.Add(Config.Name, Config);
  end;
end;

procedure TNConfigFile.CopyFromConfig(Config: TNConfigFile);
var
  I: Long;
begin
  if Config <> nil then
  begin
    // adds all Config's values
    for I := 0 to Config.FValues.Count - 1 do
      FValues.Add(Config.FValues.Keys[I], Config.FValues.Words[I]);
    // adds all Config's cloned nodes (recursively)
    for I := 0 to Config.FNodes.Count - 1 do
      AddConfig(TNConfigFile(Config.FNodes.Objects[I]).Name,
        TNConfigFile(Config.FNodes.Objects[I]).Clone);
  end;
end;

function TNConfigFile.Clone: TNConfigFile;
var
  I: Long;
begin
  Result := TNConfigFile.Create;
  Result.Name := Self.Name;
  for I := 0 to FValues.Count - 1 do
    Result.FValues.Add(FValues.Keys[I], FValues.Words[I]);
  for I := 0 to FNodes.Count - 1 do
    Result.AddConfig(TNConfigFile(FNodes.Objects[I]).Name,
      TNConfigFile(FNodes.Objects[I]).Clone);
end;

procedure TNConfigFile.Clear;
var
  I: Long;
begin
  for I := 0 to FNodes.Count - 1 do
  begin
    FNodes.Objects[I].Free;
    FNodes.Objects[I] := nil;
  end;
  FNodes.Clear;
  FValues.Clear;
end;

function TNConfigFile.FindSubNode(const NodeName: string): TNConfigFile;
var
  I: Long;
  SubName, SubNode: string;
begin
  Result := nil;
  I := PosNoCase(sConfigPathDelim, NodeName);
  if I > 0 then
  begin
    // NodeName is subnodes
    SubName := Copy(NodeName, 1, I - 1);
    SubNode := Copy(NodeName, I + 1, MaxLong);
    if FNodes.Find(SubName, I) then
      Result := TNConfigFile(FNodes.Objects[I]).FindSubNode(SubNode);
  end
  else
    // NodeName is current node's child
    if FNodes.Find(NodeName, I) then
      Result := TNConfigFile(FNodes.Objects[I]);
end;

procedure TNConfigFile.LoadFromFile(const FileName: string);
var
  F: TNBinaryFile;
begin
  try
    F := Vfs.Open(FileName, fdRead);
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'LoadFromFile', E.Message), E.ClassName);
      Exit;
    end;
  end;
  if F <> nil then
  begin
    LoadFromStream(F.Stream);
    F.Free;
  end;
end;

procedure TNConfigFile.LoadFromStream(Stream: TNStream);
var
  Driver: TNAbstractConfigDriver;
  Key, Value: string;
begin
  Clear;
  try
    Driver := GetDriverClass(Storage).Create(Stream, dmRead);
    // read and add all values until end of stream
    while not Driver.IsEnd do
    begin
      Driver.ReadValue(Key, Value);
      WriteString(Key, Value);
    end;
    Driver.Free;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'LoadFromStream', 'Error when loading ' +
        'configuration data from stream.' + sLineBreak + sExceptMsg +
        E.Message),
        E.ClassName);
  end;
end;

procedure TNConfigFile.SaveToFile(const FileName: string);
var
  F: TNBinaryFile;
begin
  try
    F := Vfs.Open(FileName, fdCreate);
  except
    on E: Exception do
    begin
      ShowError(ErrorStr(Self, 'SaveToFile', E.Message), E.ClassName);
      Exit;
    end;
  end;
  if F <> nil then
  begin
    SaveToStream(F.Stream);
    F.Free;
  end;
end;

procedure TNConfigFile.SaveToStream(Stream: TNStream);
var
  Driver: TNAbstractConfigDriver;
begin
  try
    Driver := GetDriverClass(Storage).Create(Stream, dmWrite);
    SaveAsNode(Driver);
    Driver.Free;
  except
    on E: Exception do
      ShowError(ErrorStr(Self, 'SaveToStream', 'Error when saving ' +
        'configuration data to stream.' + sLineBreak + sExceptMsg + E.Message),
        E.ClassName);
  end;
end;

end.

